import logging

import lief

logger = logging.getLogger(__file__)


def setup_logger(verbose: bool) -> None:
    # Disable logging from `lief`
    lief.logging.disable()

    if verbose:
        log_level = logging.DEBUG
    else:
        log_level = logging.INFO
    logger.setLevel(log_level)

    # Create a console handler with a higher log level
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(log_level)
    stream_handler.setFormatter(CustomFormatter())

    logger.addHandler(stream_handler)


class CustomFormatter(logging.Formatter):

    grey = "\x1b[38;20m"
    green = "\x1b[1;32m"
    yellow = "\x1b[33;20m"
    red = "\x1b[31;20m"
    bold_red = "\x1b[31;1m"
    reset = "\x1b[0m"
    format_problem_str = "%(levelname)s - %(message)s"

    FORMATS = {
        logging.DEBUG: grey + "%(levelname)s - %(message)s" + reset,
        logging.INFO: green + "%(levelname)s" + reset + " - %(message)s",
        logging.WARNING: yellow + format_problem_str + reset,
        logging.ERROR: red + format_problem_str + reset,
        logging.CRITICAL: bold_red + format_problem_str + reset
    }

    def format(self, record: logging.LogRecord) -> str:
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)



from dataclasses import dataclass
from typing import Optional, Self

import miasm.expression.expression as m2_expr
from miasm.analysis.binary import Container
from miasm.analysis.machine import Machine
from miasm.core import parse_asm
from miasm.core.asmblock import disasmEngine, asm_resolve_final
from miasm.core.interval import interval
from miasm.core.locationdb import LocationDB
from miasm.ir.ir import Lifter

MiasmFunctionInterval = interval


@dataclass
class MiasmContext:
    loc_db: LocationDB
    container: Container
    machine: Machine
    mdis: disasmEngine
    lifter: Lifter

    @classmethod
    def from_binary_file(cls, target_binary_path: str) -> Self:
        """
        Initialize our Miasm context from a binary file.
        """
        loc_db = LocationDB()
        with open(target_binary_path, 'rb') as target_binary:
            container = Container.from_stream(target_binary, loc_db)
        machine = Machine(container.arch)
        assert machine.dis_engine is not None

        mdis = machine.dis_engine(container.bin_stream, loc_db=loc_db)
        lifter = machine.lifter(loc_db)

        return cls(loc_db, container, machine, mdis, lifter)

    @property
    def arch(self) -> str:
        return str(self.machine.name)


def expr_int_to_int(expr: m2_expr.ExprInt) -> int:
    int_size = expr.size
    is_signed = expr.arg >= 2**(int_size - 1)
    if is_signed:
        result = -(2**int_size - expr.arg)
    else:
        result = expr.arg

    return int(result)


def generate_code_redirect_patch(miasm_ctx: MiasmContext, src_addr: int, dst_addr: int) -> tuple[int, bytes]:
    """
    Generate a single-block AsmCFG with a JMP from `src_addr` to `dst_addr` and return the result patch.
    """
    X86_JMP_INSTRUCTION = "JMP"

    # Generate a single-block AsmCFG with a JMP to the simplified version
    original_loc_str = f"loc_{src_addr:x}"
    jmp_unmut_instr_str = f"{original_loc_str}:\n{X86_JMP_INSTRUCTION} 0x{dst_addr:x}"
    jmp_unmut_asmcfg = parse_asm.parse_txt(miasm_ctx.mdis.arch, miasm_ctx.mdis.attrib, jmp_unmut_instr_str,
                                           miasm_ctx.mdis.loc_db)

    # Unpin loc_key if it's pinned
    original_loc = miasm_ctx.loc_db.get_offset_location(src_addr)
    if original_loc is not None:
        miasm_ctx.loc_db.unset_location_offset(original_loc)

    # Relocate the newly created block and generate machine code
    original_loc = miasm_ctx.loc_db.get_name_location(original_loc_str)
    miasm_ctx.loc_db.set_location_offset(original_loc, src_addr)
    jmp_patches = asm_resolve_final(miasm_ctx.mdis.arch, jmp_unmut_asmcfg)
    jmp_patch: Optional[tuple[int, bytes]] = next(iter(jmp_patches.items()), None)
    assert jmp_patch is not None

    return jmp_patch



from miasm.core.asmblock import AsmCFG

import themida_unmutate.symbolic_execution.x86 as symex_x86
from themida_unmutate.miasm_utils import MiasmContext, MiasmFunctionInterval


def disassemble_and_simplify_functions(miasm_ctx: MiasmContext,
                                       mutated_func_addrs: list[int]) -> list[tuple[AsmCFG, MiasmFunctionInterval]]:
    """
    Disassemble mutated functions, simplify their `AsmCFG` and return them.
    """
    match miasm_ctx.arch:
        case "x86_64":
            return symex_x86.disassemble_and_simplify_functions(miasm_ctx, mutated_func_addrs)

        case _:
            raise NotImplementedError("Unsupported architecture")



import itertools
from typing import Optional, Union

import miasm.arch.x86.arch as x86_arch
import miasm.expression.expression as m2_expr
from miasm.core.asmblock import AsmCFG, disasmEngine
from miasm.core.cpu import instruction
from miasm.core.interval import interval
from miasm.ir.symbexec import SymbolicExecutionEngine

from themida_unmutate.logging import logger
from themida_unmutate.miasm_utils import MiasmContext, MiasmFunctionInterval, expr_int_to_int

AMD64_PTR_SIZE = 64
X86_BINARY_OPS_MAPPING = {
    "+": "ADD",
    "&": "AND",
    "|": "OR",
    "^": "XOR",
    "a>>": "SAR",
    ">>": "SHR",
    "<<": "SHL",
    ">>>": "ROR",
    "<<<": "ROL",
}
AMD64_SLICES_MAPPING = {v: k for k, v in x86_arch.replace_regs64.items()}
AMD64_SP_REG = "RSP"
AMD64_IP_REG = "RIP"

MiasmIRAssignment = tuple[m2_expr.Expr, m2_expr.Expr]


def disassemble_and_simplify_functions(miasm_ctx: MiasmContext,
                                       mutated_func_addrs: list[int]) -> list[tuple[AsmCFG, MiasmFunctionInterval]]:
    """
    Disassemble mutated functions, simplify their `AsmCFG` and return them.
    """
    # Iterate through functions, disassemble and simplify them
    simplified_func_asmcfgs: list[tuple[AsmCFG, MiasmFunctionInterval]] = []
    for mutated_code_addr in mutated_func_addrs:
        logger.info("Simplifying function at 0x%x..." % mutated_code_addr)

        # Disassemble function
        asm_cfg = miasm_ctx.mdis.dis_multiblock(mutated_code_addr)
        # Compute function's interval (this is needed when rewriting the binary
        # in-place)
        original_func_interval: MiasmFunctionInterval = interval(blk.get_range() for blk in asm_cfg.blocks)

        # Lift assembly to IR
        ir_cfg = miasm_ctx.lifter.new_ircfg_from_asmcfg(asm_cfg)

        # Process IR basic blocks
        for loc_key, ir_block in ir_cfg.blocks.items():
            logger.debug("%s:" % str(loc_key))
            asm_block = asm_cfg.loc_key_to_block(loc_key)
            if asm_block is None:
                # Some instructions such `idiv` generate multiple IR basic blocks from a single asm instruction, so we
                # skip these
                continue

            relevant_assignblks = ir_block.assignblks[:-1]
            relevant_blk_count = len(relevant_assignblks)
            # No relevant instruction
            # -> unmutated, branching instruction -> keep as is
            if relevant_blk_count == 0:
                logger.debug(ir_block.assignblks[0].instr)
                continue

            # Only 1 or 2 relevant instructions
            # -> unmutated, no junk code -> no action needed -> keep first instruction as is
            if relevant_blk_count <= 2:
                logger.debug(ir_block.assignblks[0].instr)
                relocatable_instr = fix_rip_relative_instruction(asm_cfg, ir_block.assignblks[0].instr)
                # Note(ergrelet): reset the instruction's additional info to avoid
                # certain assembling issues where instruction prefixes are mixed
                # in an illegal way.
                relocatable_instr.additional_info = x86_arch.additional_info()
                relocatable_instr.additional_info.g1.value = 0
                relocatable_instr.additional_info.g2.value = 0

                asm_block.lines[0] = relocatable_instr
                continue

            reference_sb = SymbolicExecutionEngine(miasm_ctx.lifter)
            for assign_block in relevant_assignblks:
                reference_sb.eval_updt_assignblk(assign_block)
                # Forget dead stack slots
                reference_sb.del_mem_above_stack(miasm_ctx.lifter.sp)

            # Strip FLAGS register (as these are trashed by the mutation)
            strip_sym_flags(reference_sb)

            # More than 2 instructions but a single instruction replicates the symbolic state
            #   -> unmutated, junk code inserted -> keep the one instruction as is
            block_simplified = False
            for assignblk_subset in itertools.combinations(relevant_assignblks, 1):
                sb = SymbolicExecutionEngine(miasm_ctx.lifter)

                for assign_block in assignblk_subset:
                    sb.eval_updt_assignblk(assign_block)
                reference_sb.del_mem_above_stack(miasm_ctx.lifter.sp)

                # Check if instruction replicates the symbolic state
                if reference_sb.get_state() == sb.get_state():
                    for a in assignblk_subset:
                        logger.debug(a.instr)
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, a.instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                    block_simplified = True
                    break
            if block_simplified:
                continue

            # More than 2 instructions but no single instruction replicates the symbolic state
            #   -> mutated, junk code inserted -> try to "synthetize" instruction manually
            modified_variables = dict(reference_sb.modified())
            match len(modified_variables):
            # No assignment block: RET, JMP
                case 0:
                    # Keep only the last instruction
                    logger.debug(asm_block.lines[-1])
                    asm_block.lines = [asm_block.lines[-1]]
                    continue

                # 1 assignment block: MOV, XCHG, n-ary operators
                case 1:
                    ir_assignment = next(iter(modified_variables.items()))
                    dst, value = normalize_ir_assigment(ir_assignment)
                    match type(value):
                        case m2_expr.ExprId | m2_expr.ExprMem | m2_expr.ExprInt | m2_expr.ExprSlice:
                            # Assignation
                            # -> MOV
                            match type(dst):
                                case m2_expr.ExprId | m2_expr.ExprMem | m2_expr.ExprSlice:
                                    original_instr = handle_mov(miasm_ctx.mdis, dst, value)
                                    if original_instr is not None:
                                        # Update block asm block
                                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                                        continue

                        case m2_expr.ExprOp:
                            # N-ary operation on native-size registers
                            # -> ADD/SUB/INC/DEC/AND/OR/XOR/NEG/NOT/ROL/ROR/SAR/SHL/SHR
                            original_instr = handle_nary_op(miasm_ctx.mdis, dst, value)
                            if original_instr is not None:
                                # Update block asm block
                                relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                                asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                                continue

                        case m2_expr.ExprCompose:
                            # MOV, XCHG on single register or n-ary operation on lower-sized registers
                            original_instr = handle_compose(miasm_ctx.mdis, dst, value)
                            if original_instr is not None:
                                # Update block asm block
                                relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                                asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                                continue

                # 2 assignment blocks
                # -> PUSH, POP, XCHG, `SUB RSP, X`
                case 2:
                    modified_variables_iter = iter(modified_variables.items())
                    assignblk1 = next(modified_variables_iter)
                    assignblk2 = next(modified_variables_iter)

                    # PUSH
                    original_instr = handle_push(miasm_ctx.mdis, assignblk1, assignblk2)
                    if original_instr is not None:
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                        continue
                    # POP
                    original_instr = handle_pop(miasm_ctx.mdis, assignblk1, assignblk2)
                    if original_instr is not None:
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                        continue
                    # XCHG
                    original_instr = handle_xchg(miasm_ctx.mdis, assignblk1, assignblk2)
                    if original_instr is not None:
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                        continue

                    # `SUB RSP, X`
                    original_instr = handle_sub_rsp(miasm_ctx.mdis, modified_variables)
                    if original_instr is not None:
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                        continue

                # More than 2 assignment blocks
                # -> `SUB RSP, X`
                case _:
                    original_instr = handle_sub_rsp(miasm_ctx.mdis, modified_variables)
                    if original_instr is not None:
                        # Update block asm block
                        relocatable_instr = fix_rip_relative_instruction(asm_cfg, original_instr)
                        asm_block.lines = [relocatable_instr, asm_block.lines[-1]]
                        continue

            logger.debug(modified_variables)
            logger.warning("Unsupported instruction or unmutated block found. "
                           "Block will be kept as is.")

        simplified_func_asmcfgs.append((asm_cfg, original_func_interval))

    return simplified_func_asmcfgs


def strip_sym_flags(symex: SymbolicExecutionEngine) -> None:
    symex.apply_change(m2_expr.ExprId("zf", 1), m2_expr.ExprId("zf", 1))
    symex.apply_change(m2_expr.ExprId("nf", 1), m2_expr.ExprId("nf", 1))
    symex.apply_change(m2_expr.ExprId("pf", 1), m2_expr.ExprId("pf", 1))
    symex.apply_change(m2_expr.ExprId("cf", 1), m2_expr.ExprId("cf", 1))
    symex.apply_change(m2_expr.ExprId("of", 1), m2_expr.ExprId("of", 1))
    symex.apply_change(m2_expr.ExprId("af", 1), m2_expr.ExprId("af", 1))


def handle_mov(
    mdis: disasmEngine,
    dst: m2_expr.Expr,
    value: Union[m2_expr.ExprId, m2_expr.ExprMem, m2_expr.ExprInt, m2_expr.ExprSlice],
    zero_extended: bool = False,
) -> Optional[instruction]:
    if zero_extended:
        # Note(ergrelet): in x86, MOVZX only takes r/m8 and r/m16 operands
        if value.size <= 16:
            original_instr_str = f"MOVZX {ir_to_asm_str(dst)}, {ir_to_asm_str(value)}"
        else:
            # For other sizes, we simply use MOV and use a subregister as the destination
            if value.is_slice() and value.start == 0:
                dst = m2_expr.ExprSlice(dst, value.start, value.stop)
            original_instr_str = f"MOV {ir_to_asm_str(dst)}, {ir_to_asm_str(value)}"
    else:
        original_instr_str = f"MOV {ir_to_asm_str(dst)}, {ir_to_asm_str(value)}"
    original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
    logger.debug(original_instr)
    return original_instr


def handle_nary_op(mdis: disasmEngine, dst: m2_expr.Expr, op_expr: m2_expr.ExprOp) -> Optional[instruction]:
    match op_expr.op:
    # ADD, SUB, INC, DEC, NEG (we treat this separately from other binary operations)
    # Note(ergrelet): SUB is lifted as ADD(VAL1, -VAL2), NEG is lifted as ADD(0, -VAL)
        case "+":
            return handle_add_operation(mdis, dst, op_expr)
        # AND, OR, XOR, NOT, SAR, SHR, SHL, ROR, ROL
        # Note(ergrelet): NOT is lifted as XOR(VAL, -1)
        case "&" | "|" | "^" | "a>>" | ">>" | "<<" | ">>>" | "<<<":
            return handle_binary_operation(mdis, dst, op_expr)
        case _:
            return None


def handle_compose(mdis: disasmEngine, dst: m2_expr.Expr, compose_expr: m2_expr.ExprCompose) -> Optional[instruction]:
    inner_value_expr = compose_expr.args[0]

    # Match exprs of the form: '{RAX[0:32] + 0x1 0 32, 0x0 32 64}' (zero extension)
    is_zero_extension = len(compose_expr.args) == 2 and \
        compose_expr.args[1].is_int() and compose_expr.args[1].arg == 0
    # Match exprs of the form: '{RDX[0:8] + 0x1 0 8, RDX[8:64] 8 64}' (subregister)
    is_subregister_assign = len(compose_expr.args) == 2 and \
        compose_expr.args[1].is_slice() and \
        compose_expr.args[1].arg.is_id() and dst == compose_expr.args[1].arg
    if is_zero_extension or is_subregister_assign:
        match type(inner_value_expr):
            case m2_expr.ExprSlice:
                return handle_mov(mdis, dst, inner_value_expr, is_zero_extension)
            case m2_expr.ExprOp:
                return handle_nary_op(mdis, dst, inner_value_expr)

    # Handle XCHG cases where DST and SRC are subregisters of the same register
    # Match exprs of the form: '{RCX[8:16] 0 8, RCX[0:8] 8 16, RCX[16:64] 16 64}' (subregister swap)
    is_subregister_swap = len(compose_expr.args) == 3 and \
        all(map(lambda expr: expr.is_slice() and expr.arg.is_id() and expr.arg == dst,
                compose_expr.args))
    if is_subregister_swap:
        compose_lower_part, compose_mid_part, _ = compose_expr.args
        # 8-bit subregisters swap
        if compose_lower_part.size == compose_mid_part.size == 8:
            return handle_xchg(mdis, (compose_lower_part, compose_mid_part), (compose_mid_part, compose_lower_part))

    return None


def handle_add_operation(mdis: disasmEngine, dst: m2_expr.Expr, op_expr: m2_expr.ExprOp) -> Optional[instruction]:
    lhs = op_expr.args[0]
    rhs = op_expr.args[1]

    # Sub
    if rhs.is_op() and rhs.op == "-":
        if len(rhs.args) == 1 and not rhs.args[0].is_op():
            # DST = DST + (-RHS)
            if dst == lhs:
                original_instr_str = f"SUB {ir_to_asm_str(dst)}, {ir_to_asm_str(rhs.args[0])}"
                original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
                logger.debug(original_instr)
                return original_instr
            # DST = DST[0:XX] + (-RHS)
            if is_a_slice_of(lhs, dst):
                dst = m2_expr.ExprSlice(dst, lhs.start, lhs.stop)
                original_instr_str = f"SUB {ir_to_asm_str(dst)}, {ir_to_asm_str(rhs.args[0])}"
                original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
                logger.debug(original_instr)
                return original_instr

    # Sub
    if lhs.is_op() and lhs.op == "-":
        if len(lhs.args) == 1 and not lhs.args[0].is_op():
            # DST = (-LHS) + DST
            if dst == rhs:
                original_instr_str = f"SUB {ir_to_asm_str(dst)}, {ir_to_asm_str(lhs.args[0])}"
                original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
                logger.debug(original_instr)
                return original_instr
            # DST = (-LHS) + DST[0:XX]
            if is_a_slice_of(rhs, dst):
                dst = m2_expr.ExprSlice(dst, rhs.start, rhs.stop)
                original_instr_str = f"SUB {ir_to_asm_str(dst)}, {ir_to_asm_str(lhs.args[0])}"
                original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
                logger.debug(original_instr)
                return original_instr

    # TODO: handle NEG?

    # Add (regular binary operations)
    return handle_binary_operation(mdis, dst, op_expr)


def handle_binary_operation(mdis: disasmEngine, dst: m2_expr.Expr, op_expr: m2_expr.ExprOp) -> Optional[instruction]:
    op_asm_str = X86_BINARY_OPS_MAPPING.get(op_expr.op)
    if op_asm_str is None:
        # Unsupported operation
        return None

    lhs = op_expr.args[0]
    rhs = op_expr.args[1]
    if not lhs.is_op() and not rhs.is_op():
        # DST = OP(DST, RHS)
        if dst == lhs:
            original_instr_str = f"{op_asm_str} {ir_to_asm_str(dst)}, {ir_to_asm_str(rhs)}"
            original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
            logger.debug(original_instr)
            return original_instr
        # DST = OP(LHS, DST)
        if dst == rhs:
            original_instr_str = f"{op_asm_str} {ir_to_asm_str(dst)}, {ir_to_asm_str(lhs)}"
            original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
            logger.debug(original_instr)
            return original_instr
        # DST = OP(DST[0:XX], RHS)
        if is_a_slice_of(lhs, dst):
            dst = m2_expr.ExprSlice(dst, lhs.start, lhs.stop)
            original_instr_str = f"{op_asm_str} {ir_to_asm_str(dst)}, {ir_to_asm_str(rhs)}"
            original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
            logger.debug(original_instr)
            return original_instr
        # DST = OP(LHS, DST[0:XX])
        if is_a_slice_of(rhs, dst):
            dst = m2_expr.ExprSlice(dst, rhs.start, rhs.stop)
            original_instr_str = f"{op_asm_str} {ir_to_asm_str(dst)}, {ir_to_asm_str(lhs)}"
            original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
            logger.debug(original_instr)
            return original_instr

    return None


def handle_xchg(mdis: disasmEngine, ir_assignment1: MiasmIRAssignment,
                ir_assignment2: MiasmIRAssignment) -> Optional[instruction]:
    norm_assignment1 = normalize_ir_assigment(ir_assignment1)
    norm_assignment2 = normalize_ir_assigment(ir_assignment2)

    # Handle most XCHG cases where DST and SRC are swapped and aren't part of
    # the same register
    if norm_assignment1[0] == norm_assignment2[1] and \
            norm_assignment2[0] == norm_assignment1[1]:
        original_instr_str = f"XCHG {ir_to_asm_str(norm_assignment2[0])}, {ir_to_asm_str(norm_assignment2[1])}"
        original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
        logger.debug(original_instr)
        return original_instr

    return None


def handle_push(mdis: disasmEngine, ir_assignment1: MiasmIRAssignment,
                ir_assignment2: MiasmIRAssignment) -> Optional[instruction]:
    rsp_decrement_op = m2_expr.ExprOp("+", m2_expr.ExprId(AMD64_SP_REG, AMD64_PTR_SIZE),
                                      m2_expr.ExprInt(0xFFFFFFFFFFFFFFF8, AMD64_PTR_SIZE))
    is_rsp_decremented = ir_assignment1[1] == rsp_decrement_op
    is_dst1_rsp = ir_assignment1[0].is_id() and ir_assignment1[0].name == AMD64_SP_REG
    is_dst2_on_stack = ir_assignment2[0].is_mem() and \
                       ir_assignment2[0].ptr == rsp_decrement_op

    if is_dst1_rsp and is_rsp_decremented and is_dst2_on_stack:
        original_instr_str = f"PUSH {ir_to_asm_str(ir_assignment2[1])}"
        original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
        logger.debug(original_instr)
        return original_instr

    return None


def handle_pop(mdis: disasmEngine, ir_assignment1: MiasmIRAssignment,
               ir_assignment2: MiasmIRAssignment) -> Optional[instruction]:
    rsp_increment_op = m2_expr.ExprOp("+", m2_expr.ExprId(AMD64_SP_REG, AMD64_PTR_SIZE),
                                      m2_expr.ExprInt(0x8, AMD64_PTR_SIZE))
    is_rsp_incremented = ir_assignment2[1] == rsp_increment_op
    is_value1_on_stack = ir_assignment1[1].is_mem() and \
                         ir_assignment1[1].ptr.is_id() and ir_assignment1[1].ptr.name == AMD64_SP_REG
    is_dst2_rsp = ir_assignment2[0].is_id() and \
                  ir_assignment2[0].name == AMD64_SP_REG

    if is_value1_on_stack and is_rsp_incremented and is_dst2_rsp:
        original_instr_str = f"POP {ir_to_asm_str(ir_assignment1[0])}"
        original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
        logger.debug(original_instr)
        return original_instr

    return None


# Note(ergrelet): `SUB RSP, X` is a special case where there might be some residual
# constraints that the symbolic execution cannot discard, because allocated stack
# slots are live and values might have been assigned to some of them by the
# inserted junk code.
# We thus have to differentiate legit PUSH-like instructions from junked
# `SUB RSP, X` instructions.
def handle_sub_rsp(mdis: disasmEngine, assign_blk: dict[m2_expr.Expr, m2_expr.Expr]) -> Optional[instruction]:

    def is_sub_rsp_expr(expr: m2_expr.Expr) -> bool:
        """
        This matches OP expressions of the form "RSP - X"
        """
        if not expr.is_op():
            return False

        is_binary_add = expr.op == "+" and len(expr.args) == 2
        if not is_binary_add:
            return False

        # One of the operands must be RSP
        rsp_in_expr = any(map(lambda arg: arg.is_id() and arg.name == AMD64_SP_REG, expr.args))
        # The other operand must be a negative integer
        neg_int_in_expr = any(map(lambda arg: arg.is_int() and expr_int_to_int(arg) < 0, expr.args))

        return rsp_in_expr and neg_int_in_expr

    def is_sub_rsp_blk(ir_assignment: MiasmIRAssignment) -> bool:
        """
        This matches assignments of the form "RSP - X"
        """
        dst, src = ir_assignment
        if not dst.is_id() or not src.is_op():
            return False

        # Destination must be RSP
        dst_is_rsp = dst.name == AMD64_SP_REG
        return dst_is_rsp and is_sub_rsp_expr(src)

    # Check if a SUB operation is applied to RSP
    sub_rsp_blk = next(filter(is_sub_rsp_blk, assign_blk.items()), None)
    if sub_rsp_blk is not None:
        # Extract allocated size
        allocated_window = (0, expr_int_to_int(sub_rsp_blk[1].args[1]))
        # Check if all allocated slots have been written to or not. This
        # assumes no PUSH-like x86 instruction allocates more slots than needed to
        # write the data it pushes to the stack
        #
        # FIXME: we're not checking individual stack slots but rather
        # the outter boundaries from used stack slots. It would be better
        # to track accesses to each stack slot separately.
        used_window = [0, 0]
        for dst in assign_blk.keys():
            if dst.is_mem() and is_sub_rsp_expr(dst.ptr):
                dst_offset_in_stack = expr_int_to_int(dst.ptr.args[1])
                dst_size_in_bytes = dst.size // 8
                used_window[0] = max(used_window[0], dst_offset_in_stack + dst_size_in_bytes)
                used_window[1] = min(used_window[1], dst_offset_in_stack)
        if tuple(used_window) == allocated_window:
            # All allocated slots are used, must be a PUSH-like instruction
            return None

        original_instr_str = f"SUB {AMD64_SP_REG}, {-allocated_window[1]}"
        original_instr = mdis.arch.fromstring(original_instr_str, mdis.loc_db, mdis.attrib)
        logger.debug(original_instr)
        return original_instr

    return None


def ir_to_asm_str(expr: m2_expr.Expr) -> str:
    match type(expr):
        case m2_expr.ExprMem:
            return mem_ir_to_asm_str(expr)
        case m2_expr.ExprSlice:
            return slice_ir_to_asm_str(expr)
        case _:
            return str(expr)


def mem_ir_to_asm_str(mem_expr: m2_expr.ExprMem) -> str:
    match mem_expr.size:
        case 128 | 64 | 32 | 16 | 8:
            mem_prefix = x86_arch.SIZE2MEMPREFIX[mem_expr.size]
            return f"{mem_prefix} PTR [{mem_expr.ptr}]"
        case _:
            raise ValueError("Invalid ExprMem size")


def slice_ir_to_asm_str(slice_expr: m2_expr.ExprSlice) -> str:
    match type(slice_expr.arg):
        case m2_expr.ExprId:
            # Slice of a register
            return str(AMD64_SLICES_MAPPING[slice_expr])
        case _:
            return str(slice_expr)


def normalize_ir_assigment(ir_assignment: MiasmIRAssignment) -> MiasmIRAssignment:
    """
    Normalize IR assignments by transforming `ExprCompose`s in SRC into
    `ExprSlice`s in DST when appropriate.
    This allows us to properly detect assigments made to subregisters
    (e.g., `EAX`, `AX`, `AL`, `AH`).
    """
    dst, src = ir_assignment

    # Match ExprId(X) = ExprCompose(Y)
    if dst.is_id() and src.is_compose():
        match len(src.args):
        # 2-way compose (e.g., `EAX`, `AX`, `AL`)
            case 2:
                compose_lower_part, compose_upper_part = src.args
                # If upper bits from DST are kept, it means we're dealing with
                # a 16-bit or 8-bit subregister
                if compose_upper_part.arg == dst and \
                        compose_upper_part.start == compose_lower_part.size and \
                        compose_upper_part.stop == dst.size:
                    # DST -> DST[0:X]
                    new_dst = m2_expr.ExprSlice(dst, 0, compose_lower_part.size)
                    # Concat(SRC[0:X], DST[X:]) -> SRC[0:X]
                    new_src = compose_lower_part
                    return (new_dst, new_src)

                # If upper bits are zeroed out, it means mean we're dealing with
                # a 32-bit subregister
                upper_zero_bits = m2_expr.ExprInt(0, AMD64_PTR_SIZE // 2)
                if compose_upper_part == upper_zero_bits:
                    # DST -> DST[0:32]
                    new_dst = m2_expr.ExprSlice(dst, 0, compose_lower_part.size)
                    # Concat(SRC, DST[X:]) -> SRC
                    new_src = compose_lower_part

                    return (new_dst, new_src)

            # 3-way compose (e.g., `AH`, `BH`)
            case 3:
                compose_lower_part, compose_mid_part, compose_upper_part = src.args
                # If lower and upper bits from DST are kept, it means we're
                # dealing with a subregister
                if compose_lower_part.arg == dst and \
                        compose_lower_part.start == 0 and \
                        compose_upper_part.arg == dst and \
                        compose_upper_part.stop == dst.size:
                    # DST -> DST[X:Y]
                    new_dst = m2_expr.ExprSlice(dst, compose_lower_part.size,
                                                compose_lower_part.size + compose_mid_part.size)
                    # Concat(DST[0:X], SRC, DST[Y:]) -> SRC
                    new_src = compose_mid_part
                    return (new_dst, new_src)

    return ir_assignment


def is_a_slice_of(slice_expr: m2_expr.Expr, expr: m2_expr.Expr) -> bool:
    return slice_expr.is_slice() and slice_expr.arg == expr  # type:ignore


# Fix RIP relative instructions to make them relocatable
def fix_rip_relative_instruction(asmcfg: AsmCFG, instr: instruction) -> instruction:
    rip = m2_expr.ExprId(AMD64_IP_REG, AMD64_PTR_SIZE)
    # Note(ergrelet): see https://github.com/cea-sec/miasm/issues/1258#issuecomment-645640366
    # for more information on what the '_' symbol is used for.
    new_next_addr_card = m2_expr.ExprLoc(asmcfg.loc_db.get_or_create_name_location('_'), AMD64_PTR_SIZE)
    for i, arg in enumerate(instr.args):
        if rip in arg:
            assert instr.offset is not None and instr.l is not None
            next_instr_addr = m2_expr.ExprInt(instr.offset + instr.l, AMD64_PTR_SIZE)
            fix_dict = {rip: rip + next_instr_addr - new_next_addr_card}
            instr.args[i] = arg.replace_expr(fix_dict)

    return instr



from argparse import ArgumentParser, Namespace

from themida_unmutate.logging import setup_logger, logger
from themida_unmutate.miasm_utils import MiasmContext
from themida_unmutate.rebuilding import rebuild_simplified_binary
from themida_unmutate.symbolic_execution import disassemble_and_simplify_functions
from themida_unmutate.unwrapping import unwrap_functions


def entry_point() -> None:
    # Parse command-line arguments
    args = parse_arguments()

    # Setup logging
    setup_logger(args.verbose)

    # Setup disassembler and lifter
    miasm_ctx = MiasmContext.from_binary_file(args.protected_binary)

    # Resolve mutated functions' addresses if needed
    protected_func_addrs = list(map(lambda addr: int(addr, 0), args.addresses))
    if not args.no_trampoline:
        logger.info("Resolving mutated's functions' addresses...")
        mutated_func_addrs = unwrap_functions(miasm_ctx, protected_func_addrs)
    else:
        # No trampolines to take care of, use target addresses directly
        mutated_func_addrs = protected_func_addrs

    # Disassemble mutated functions and simplify them
    logger.info("Deobfuscating mutated functions...")
    simplified_func_asmcfgs = disassemble_and_simplify_functions(miasm_ctx, mutated_func_addrs)

    # Map protected functions' addresses to their corresponding simplified `AsmCFG`
    func_addr_to_simplified_cfg = {
        protected_func_addrs[i]: asm_cfg
        for i, asm_cfg in enumerate(simplified_func_asmcfgs)
    }

    # Rewrite the protected binary with simplified functions
    logger.info("Rebuilding binary file...")
    rebuild_simplified_binary(miasm_ctx, func_addr_to_simplified_cfg, args.protected_binary, args.output,
                              args.reassemble_in_place)

    logger.info("Done! You can find your deobfuscated binary at '%s'" % args.output)


def parse_arguments() -> Namespace:
    """
    Parse command-line arguments.
    """
    parser = ArgumentParser(description="Automatic deobfuscation tool for Themida's mutation-based protection")
    parser.add_argument("protected_binary", help="Protected binary path")
    parser.add_argument("-a", "--addresses", nargs='+', help="Addresses of the functions to deobfuscate", required=True)
    parser.add_argument("-o", "--output", help="Output binary path", required=True)
    parser.add_argument("--no-trampoline", action='store_true', help="Disable function unwrapping")
    parser.add_argument("--reassemble-in-place",
                        action='store_true',
                        help="Rewrite simplified code over the mutated code"
                        "rather than in a new code section")
    parser.add_argument("-v", "--verbose", action='store_true', help="Enable verbose logging")

    return parser.parse_args()


if __name__ == "__main__":
    entry_point()



import miasm.expression.expression as m2_expr
from miasm.ir.ir import IRCFG, Lifter
from miasm.ir.symbexec import SymbolicExecutionEngine

from themida_unmutate.logging import logger
from themida_unmutate.miasm_utils import MiasmContext, expr_int_to_int


def unwrap_functions(miasm_ctx: MiasmContext, target_function_addrs: list[int]) -> list[int]:
    """
    Resolve mutated function's addresses from original function addresses.
    """
    mutated_func_addrs: list[int] = []
    for addr in target_function_addrs:
        logger.debug("Resolving mutated code portion address for 0x%x..." % addr)
        mutated_code_addr = _resolve_mutated_code_address(miasm_ctx, addr)
        if mutated_code_addr == addr:
            raise Exception("Failure to unwrap function")

        logger.info("Function at 0x%x jumps to 0x%x" % (addr, mutated_code_addr))
        mutated_func_addrs.append(mutated_code_addr)

    return mutated_func_addrs


def _resolve_mutated_code_address(miasm_ctx: MiasmContext, target_addr: int) -> int:
    # Save `follow_call` value and set it to `True`
    saved_follow_call = miasm_ctx.mdis.follow_call
    miasm_ctx.mdis.follow_call = True
    # Disassemble trampoline
    asmcfg = miasm_ctx.mdis.dis_multiblock(target_addr)
    # Restore `follow_call` value
    miasm_ctx.mdis.follow_call = saved_follow_call
    # Lift ASM to IR
    ircfg = miasm_ctx.lifter.new_ircfg_from_asmcfg(asmcfg)

    return _resolve_mutated_portion_address(miasm_ctx.lifter, ircfg, target_addr)


def _resolve_mutated_portion_address(lifter: Lifter, ircfg: IRCFG, call_addr: int) -> int:
    # Instantiate a Symbolic Execution engine with default value for registers
    symb = SymbolicExecutionEngine(lifter)

    # Emulate until the next address cannot be resolved (`ret`, unresolved condition, etc.)
    cur_expr = symb.run_at(ircfg, call_addr)

    # First `cmp` -> eval to zero
    if not cur_expr.is_cond() or not cur_expr.cond.is_mem():
        logger.warning("Function doesn't behave as expected, considering it unmutated")
        return call_addr

    # Value if condition is evaled zero
    symb.eval_updt_expr(m2_expr.ExprAssign(cur_expr.cond, m2_expr.ExprInt(0, cur_expr.cond.size)))
    target = cur_expr.src2
    cur_expr = symb.run_at(ircfg, target)

    # Second `cmp` -> eval to zero
    if not cur_expr.is_cond() or not cur_expr.cond.is_mem():
        logger.warning("Function doesn't behave as expected, considering it unmutated")
        return call_addr

    # Value if condition is evaled zero
    symb.eval_updt_expr(m2_expr.ExprAssign(cur_expr.cond, m2_expr.ExprInt(0, cur_expr.cond.size)))
    target = cur_expr.src2
    cur_expr = symb.run_at(ircfg, target)
    if not isinstance(cur_expr, m2_expr.ExprInt):
        # If we're here, this might be a Themida 3.1.7+ trampoline, handle the additional JCC
        if not cur_expr.is_cond() or not cur_expr.cond.is_mem():
            logger.warning("Function doesn't behave as expected, considering it unmutated")
            return call_addr

        symb.eval_updt_expr(m2_expr.ExprAssign(cur_expr.cond, m2_expr.ExprInt(0, cur_expr.cond.size)))
        target = cur_expr.src2
        cur_expr = symb.run_at(ircfg, target)
        if not isinstance(cur_expr, m2_expr.ExprInt):
            logger.warning("Function doesn't behave as expected, considering it unmutated")
            return call_addr

    # This time we should have the real mutated function address
    return expr_int_to_int(cur_expr)



from typing import Optional

import lief
from miasm.core.asmblock import AsmCFG, asm_resolve_final, bbl_simplifier
from miasm.core.interval import interval

from themida_unmutate.miasm_utils import MiasmContext, MiasmFunctionInterval, generate_code_redirect_patch

NEW_SECTION_NAME = ".unmut"
NEW_SECTION_MAX_SIZE = 2**16


def rebuild_simplified_binary(
    miasm_ctx: MiasmContext,
    func_addr_to_simplified_cfg: dict[int, tuple[AsmCFG, MiasmFunctionInterval]],
    input_binary_path: str,
    output_binary_path: str,
    reassemble_in_place: bool,
) -> None:
    """
    Reassemble functions' `AsmCFG` and rewrite the input binary with simplified
    machine code.
    """
    if len(func_addr_to_simplified_cfg) == 0:
        raise ValueError("`protected_function_addrs` cannot be empty")

    if reassemble_in_place:
        __rebuild_simplified_binary_in_place(miasm_ctx, func_addr_to_simplified_cfg, input_binary_path,
                                             output_binary_path)
    else:
        __rebuild_simplified_binary_in_new_section(miasm_ctx, func_addr_to_simplified_cfg, input_binary_path,
                                                   output_binary_path)


def __rebuild_simplified_binary_in_new_section(
    miasm_ctx: MiasmContext,
    func_addr_to_simplified_cfg: dict[int, tuple[AsmCFG, MiasmFunctionInterval]],
    input_binary_path: str,
    output_binary_path: str,
) -> None:
    """
    Reassemble functions' `AsmCFG` and rewrite the input binary with simplified
    machine code in a new code section.
    """
    # Open the target binary with LIEF
    pe_obj = lief.PE.parse(input_binary_path)
    if pe_obj is None:
        raise Exception(f"Failed to parse PE '{input_binary_path}'")

    # Create a new code section
    unmut_section = lief.PE.Section([0] * NEW_SECTION_MAX_SIZE, NEW_SECTION_NAME,
                                    lief.PE.SECTION_CHARACTERISTICS.CNT_CODE.value
                                    | lief.PE.SECTION_CHARACTERISTICS.MEM_READ.value
                                    | lief.PE.SECTION_CHARACTERISTICS.MEM_EXECUTE.value)
    pe_obj.add_section(unmut_section)
    unmut_section = pe_obj.get_section(NEW_SECTION_NAME)
    unmut_section_base = pe_obj.imagebase + unmut_section.virtual_address

    # Reassemble simplified AsmCFGs
    original_to_simplified: dict[int, int] = {}
    next_min_offset_for_asm = 0
    unmut_section_patches: list[tuple[int, bytes]] = []
    for protected_func_addr, val in \
            func_addr_to_simplified_cfg.items():
        simplified_asmcfg, _ = val
        # Simplify CFG further (by merging basic blocks when possible)
        simplified_asmcfg = bbl_simplifier(simplified_asmcfg)

        # Unpin blocks to be able to relocate the whole CFG
        head = simplified_asmcfg.heads()[0]
        for ir_block in simplified_asmcfg.blocks:
            miasm_ctx.loc_db.unset_location_offset(ir_block.loc_key)

        # Relocate the function's entry block
        miasm_ctx.loc_db.set_location_offset(head, unmut_section_base + next_min_offset_for_asm)

        # Generate the simplified machine code
        new_section_patches = asm_resolve_final(
            miasm_ctx.mdis.arch,
            simplified_asmcfg,
            dst_interval=interval([(unmut_section_base + next_min_offset_for_asm,
                                    unmut_section_base + unmut_section.virtual_size - next_min_offset_for_asm)]))

        # Merge patches into the patch list
        for patch in new_section_patches.items():
            unmut_section_patches.append(patch)

        # Associate original addr to simplified addr
        original_to_simplified[protected_func_addr] = min(new_section_patches.keys())
        next_min_offset_for_asm = max(new_section_patches.keys()) - unmut_section_base + 15

    # Overwrite the new section's content
    new_section_size = next_min_offset_for_asm
    new_content = bytearray([0] * new_section_size)
    for addr, data in unmut_section_patches:
        offset = addr - unmut_section_base
        new_content[offset:offset + len(data)] = data
    unmut_section.content = memoryview(new_content)

    # Redirect functions to their simplified versions
    protected_function_addrs = func_addr_to_simplified_cfg.keys()
    unmut_jmp_patches: list[tuple[int, bytes]] = []
    for target_addr in protected_function_addrs:
        # Generate a single-block AsmCFG with a JMP to the simplified version
        simplified_func_addr = original_to_simplified[target_addr]
        unmut_jmp_patch = generate_code_redirect_patch(miasm_ctx, target_addr, simplified_func_addr)
        unmut_jmp_patches.append(unmut_jmp_patch)

    # Find the section containing the original function
    text_section = __section_from_virtual_address(pe_obj, next(iter(protected_function_addrs)))
    assert text_section is not None

    # Apply patches
    text_section_base = pe_obj.imagebase + text_section.virtual_address
    text_section_bytes = bytearray(text_section.content)
    for addr, data in unmut_jmp_patches:
        offset = addr - text_section_base
        text_section_bytes[offset:offset + len(data)] = data
    text_section.content = memoryview(text_section_bytes)

    # Invoke the builder
    builder = lief.PE.Builder(pe_obj)
    builder.build()

    # Save the result
    builder.write(output_binary_path)


def __rebuild_simplified_binary_in_place(
    miasm_ctx: MiasmContext,
    func_addr_to_simplified_cfg: dict[int, tuple[AsmCFG, MiasmFunctionInterval]],
    input_binary_path: str,
    output_binary_path: str,
) -> None:
    """
    Reassemble functions' `AsmCFG` and rewrite the input binary with simplified
    machine code by overwriting the mutated code.
    """
    # Open the target binary with LIEF
    pe_obj = lief.PE.parse(input_binary_path)
    if pe_obj is None:
        raise Exception(f"Failed to parse PE '{input_binary_path}'")

    # Reassemble simplified AsmCFGs
    original_to_simplified: dict[int, int] = {}
    unmut_patches: list[tuple[int, bytes]] = []
    for protected_func_addr, val in \
            func_addr_to_simplified_cfg.items():
        simplified_asmcfg, orignal_asmcfg_interval = val

        # Unpin blocks to be able to relocate the CFG
        head = simplified_asmcfg.heads()[0]
        for asm_block in simplified_asmcfg.blocks:
            miasm_ctx.loc_db.unset_location_offset(asm_block.loc_key)

        # Start rewriting at the first part of the interval (i.e., at the start
        # of the mutated code)
        target_addr: int = orignal_asmcfg_interval.intervals[0][0]
        # Unpin loc_key if it's pinned
        original_loc = miasm_ctx.loc_db.get_offset_location(target_addr)
        if original_loc is not None:
            miasm_ctx.loc_db.unset_location_offset(original_loc)

        # Relocate the function's entry block
        miasm_ctx.loc_db.set_location_offset(head, target_addr)

        # Generate the simplified machine code
        new_section_patches = asm_resolve_final(miasm_ctx.mdis.arch,
                                                simplified_asmcfg,
                                                dst_interval=orignal_asmcfg_interval)

        # Merge patches into the patch list
        for patch in new_section_patches.items():
            unmut_patches.append(patch)

        # Associate original addr to simplified addr
        original_to_simplified[protected_func_addr] = min(new_section_patches.keys())

    # Find Themida's section
    themida_section = __section_from_virtual_address(pe_obj, target_addr)
    assert themida_section is not None

    # Overwrite Themida's section content
    themida_section_base = pe_obj.imagebase + themida_section.virtual_address
    new_content = bytearray(themida_section.content)
    for addr, data in unmut_patches:
        offset = addr - themida_section_base
        new_content[offset:offset + len(data)] = data
    themida_section.content = memoryview(new_content)

    # Redirect functions to their simplified versions
    protected_function_addrs = func_addr_to_simplified_cfg.keys()
    unmut_jmp_patches: list[tuple[int, bytes]] = []
    for target_addr in protected_function_addrs:
        # Generate a single-block AsmCFG with a JMP to the simplified version
        simplified_func_addr = original_to_simplified[target_addr]
        unmut_jmp_patch = generate_code_redirect_patch(miasm_ctx, target_addr, simplified_func_addr)
        unmut_jmp_patches.append(unmut_jmp_patch)

    # Find the section containing the original function
    text_section = __section_from_virtual_address(pe_obj, next(iter(protected_function_addrs)))
    assert text_section is not None

    # Apply patches
    text_section_base = pe_obj.imagebase + text_section.virtual_address
    text_section_bytes = bytearray(text_section.content)
    for addr, data in unmut_jmp_patches:
        offset = addr - text_section_base
        text_section_bytes[offset:offset + len(data)] = data
    text_section.content = memoryview(text_section_bytes)

    # Invoke the builder
    builder = lief.PE.Builder(pe_obj)
    builder.build()

    # Save the result
    builder.write(output_binary_path)


def __section_from_virtual_address(lief_bin: lief.Binary, virtual_addr: int) -> Optional[lief.Section]:
    rva = virtual_addr - lief_bin.imagebase
    return __section_from_rva(lief_bin, rva)


def __section_from_rva(lief_bin: lief.Binary, rva: int) -> Optional[lief.Section]:
    for s in lief_bin.sections:
        if s.virtual_address <= rva < s.virtual_address + s.size:
            assert isinstance(s, lief.Section)
            return s

    return None
